package chat.tcp.server;

import java.io.*;
import java.net.*;

 chat.server.*;
import chat.tcp.common.TCPChatMessageTypes;

// This class acts like a client of the ChatServer application. It
// translates messages from a TCP socket into requests for the chat
// server, and translates method invocations from the server into
// TCP messages.

public class TCPChatClient extends Object 
implements ChatClient, Runnable
{
// server is the ChatServer application we are a client of
     protected ChatServer server;

// clientSock is the socket connection to the user
     protected Socket clientSock;

// inStream and outStream are Data streams for the socket. This allows
// us to send information in forms other than an array of bytes

     protected DataInputStream inStream;
     protected DataOutputStream outStream;

// clientName is the name the user wants to be known by
     protected String clientName;

     protected Thread myThread;

     public TCPChatClient(ChatServer server, Socket clientSock)
     throws IOException
     {
          this.server = server;
          this.clientSock = clientSock;

// get data streams to the socket

          inStream = new DataInputStream(
               clientSock.getInputStream());

          outStream = new DataOutputStream(
               clientSock.getOutputStream());

// The first thing that the user sends us is 
// the name they want to use
          clientName = inStream.readUTF();

// Add ourself to the server application
          server.addClient(clientName, this);
     }

// The next few methods implement a really simple messaging protocol:
// 4 byte Integer message type
// 4 byte message length
// <message length> bytes of data

// userHasEntered is called by the server whenever there's a new user
// the data part of the message is just the name of the user who has
// entered.

     public void userHasEntered(String who)
     {
          try {
// Write the message type
               outStream.writeInt(TCPChatMessageTypes.ENTER);
// Write the message length
               outStream.writeInt(who.length());
// Write the user's name
               outStream.writeBytes(who);
          } catch (Exception e) {
               server.removeClient(this);
          }
     }

// userHasLeft is called by the server whenever there's a new user
// the data part of the message is just the name of the user who has
// left.
     public void userHasLeft(String who)
     {
          try {
               outStream.writeInt(TCPChatMessageTypes.LEAVE);
               outStream.writeInt(who.length());
               outStream.writeBytes(who);
          } catch (Exception e) {
               server.removeClient(this);
          }
     }

// incomingChat is called by the server whenever someone sends a message.
// The data part of the message has three parts:
// the length of the name of the person sending the message (the
// length value itself is a 4-byte integer)
// the name of the person sending the message
// the chat message

     public void incomingChat(String who, String chat)
     {
          try {
               outStream.writeInt(TCPChatMessageTypes.CHAT);
               outStream.writeInt(who.length() + chat.length() + 4);
               outStream.writeInt(who.length());
               outStream.writeBytes(who);
               outStream.writeBytes(chat);
          } catch (Exception e) {
               server.removeClient(this);
          }
     }

// disconnect is called by the server when the client has 
// been disconnected from the server. We just close down the
// socket and stop this thread.

     public void disconnect()
     {
          try {
               clientSock.close();
          } catch (Exception e) {
          }
          stop();
     }
	 
	 // handleChatMessage reads an incoming chat message from the user and
// sends it to the server. The data part of the message is just the
// chat message itself.

     public void handleChatMessage()
     throws IOException
     {
// Get the message length
          int length = inStream.readInt();
          byte[] chatChars = new byte[length];

// Read the chat message
          inStream.readFully(chatChars);

          String message = new String(chatChars, 0);

// Send the chat message to the server
          server.sendChat(clientName, message);
     }

// If we get a message we don't understand, skip over it. That's
// why we have the message length as part of the protocol.

     public void skipMessage()
     throws IOException
     {
          int length = inStream.readInt();
          inStream.skipBytes(length);
     }

     public void run()
     {
          while (true) {
               try {

// Read the type of the next message
                    int messageType = inStream.readInt();

                    switch (messageType) {

// If it's a chat message, read it
                         case TCPChatMessageTypes.CHAT:
                              handleChatMessage();
                              break;

// For any messages whose type we don't understand, skip the message
                         default:
                              skipMessage();
                              return;
                    }
               } catch (Exception e) {
                    server.removeClient(clientName);
                    return;
               }
 }
     }
     public void start()
     {
          myThread = new Thread(this);
          myThread.start();
     }

     public void stop()
     {
          myThread.stop();
          myThread = null;
     }
}